/*

  timing.cpp

*/

#include "timing.h"
#include "gui.h"

static LARGE_INTEGER initime;
static LARGE_INTEGER oldt;
static LARGE_INTEGER pfq;
static __int64 oneframe,pfq64;

static long oldframes;
static long oldsec;

extern tGui* gui;

//

void init_timer()
{
	BOOL rv;

	// check for high precision timer

	rv=QueryPerformanceFrequency(&pfq);
	if (rv==0)
	{
		MessageBox(0,"this machine does not support rdtsc","error",MB_OK);
	}

	pfq64=pfq.QuadPart;
	oneframe=pfq64/60;

	rv=QueryPerformanceCounter(&oldt);
	if (rv==0)
	{
		MessageBox(0,"this machine does not support rdtsc","error",MB_OK);
	}

	oldframes=0;
	oldsec=0;
}

//

void trim_speed()
{
	// see how much time is elapsed
	// between this call and the last one

	LARGE_INTEGER curt;
	__int64 curt64,oldt64;
	
	QueryPerformanceCounter(&curt);

	curt64=curt.QuadPart;
	oldt64=oldt.QuadPart;

	while ((curt64-oldt64)<oneframe)
	{
		QueryPerformanceCounter(&curt);
		curt64=curt.QuadPart;
	}

	oldt=curt;
}

void start_count()
{
	QueryPerformanceCounter(&initime);

	oldframes=0;
	oldsec=0;
}

int secs_elapsed()
{
	LARGE_INTEGER curtime,freq;
	__int64 curt64,oldt64,freq64;
	int secs;
	
	QueryPerformanceCounter(&curtime);
	QueryPerformanceFrequency(&freq);

	curt64=curtime.QuadPart;
	oldt64=initime.QuadPart;
	freq64=freq.QuadPart;

	secs=(int)((curt64-oldt64)/freq64);

	return (secs);
}

//

void update_fps(int curframes,float& gfps,int& frameskip)
{
	int sec=secs_elapsed();

	if (sec>oldsec)
	{
		// one second is elapsed:
		// update oldsec, update oldframes
		// count fps

		gfps=(float)(curframes-oldframes);

		oldframes=curframes;
		oldsec=sec;

		if (gui->guiinfo.autoframeskip)
		{
			// if fps are less than 59
			// lower down the frameskip (down to 10 max)

			if (gfps<59)
			{
				frameskip++;

				if (frameskip>10) frameskip=10;
			}
			else
			{
				// try to raise the frameskip
				// if it is more than 1

				if (frameskip>1)
				{
					frameskip--;
				}
			}
		}
	}
}
