/*
**
** File: ym2151.h -- header file for software implementation of YM-2151
**						FM Operator Type-M(OPM).
**
** (C) 1997,1998 Jarek Burczynski (s0246@goblin.pjwstk.waw.pl)
** Many of the optimizing ideas by Tatsuyuki Satoh
**
** Version 2.0  November, 29th 1998
**
**
** CAUTION !!! CAUTION !!!           H E L P   W A N T E D !!!
**
** If you have some very technical information about YM2151 and you would
** like to help:
**                  - PLEASE CONTACT ME ASAP!!! -
**
**
** Big THANK YOU to Beauty Planets for making a lot of real YM2151 samples and
** also for giving additional informations about the chip. Also for the time
** spent making the samples and the speed of replying to my endless requests.
**
** I would like to thank to Shigeharu Isoda without whom this project wouldn't
** be possible to finish (or even start). He has been so kind and helped me
** by scanning his YM2151 Japanese Manual first, and then by answering MANY
** of my questions.
**
** I would like to thank to nao also - for giving me some informations about
** YM2151 and pointing me to Shigeharu.
**
** Big thanks to Aaron Giles and Chris Hardy - they made some samples of one
** of my favourite arcade games so I could compare it to my emulator. That's
** how I know there's still something I need to know about this chip.
**
** Greetings to Ishmair (thanks for the datasheet !) who inspired me to write
** my own 2151 emu by telling me "... maybe with some somes and a lot of
** luck... in theory it's difficult... " - wow that really was a challenge :)
**
*/

#ifndef _H_YM2151_
#define _H_YM2151_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

//typedef void SAMP;  /* 16-bit or 8-bit sample modes supported */
#define SAMP short int

typedef struct {      /*oscillator data */
    unsigned int phase; /*accumulated oscillator phase (in indexes)    (fixed point)*/
    unsigned int freq;  /*oscillator frequency (in indexes per sample) (fixed point)*/
    signed int LFOpm;   /*phase modulation from LFO*/

    unsigned int volume;    /*oscillator volume (fixed point)   */
    signed int attack_volume;/*used for attack phase calculations*/

    unsigned int TL;    /*Total attenuation Level*/
    unsigned int delta_AR;  /*volume delta for attack  phase (fixed point)*/
    unsigned int delta_D1R; /*volume delta for decay   phase (fixed point)*/
    unsigned int D1L;   /*when envelope reaches this level Envelope Generator goes into D2R*/
    unsigned int delta_D2R; /*volume delta for sustain phase (fixed point)*/
    unsigned int delta_RR;  /*volume delta for release phase (fixed point)*/
    unsigned int state; /*Envelope state: 4-attack(AR) 3-decay(D1R) 2-sustain(D2R) 1-release(RR) 0-off*/

    unsigned int AMSmask;   /*LFO AMS enable mask*/

    signed int * connect;   /*oscillator output direction*/

/*Begin of channel specific data*/
/*The trick : each operator number 0 contains channel specific data*/
    signed int OscilFB; /*oscillator self feedback value used only by operators 0*/
    unsigned int FeedBack;  /*feedback shift value for operators 0 in each channel*/
    unsigned int KC;    /*KC for each channel*/
    unsigned int KF;    /*KF for each channel*/
    unsigned int PMS;   /*PMS for each channel*/
    unsigned int AMS;   /*AMS for each channel*/
    unsigned int PAN;   /*bit 7-right enable, bit 6-left enable: 0-off, 1-on*/
/*End of channel specific data*/

    unsigned int key;   /*0=last key was KEYOFF, 1=last key was KEON*/

    unsigned int KS;    /*Key Scale*/
    unsigned int AR;    /*Attack rate*/
    unsigned int D1R;   /*Decay rate*/
    unsigned int D2R;   /*Sustain rate*/
    unsigned int RR;    /*Release rate*/
    unsigned int mul;   /*phase multiply*/
    unsigned int DT1;   /*DT1 index*32 */
    unsigned int DT2;   /*DT2 index*/
    unsigned int KCindex;   /*Used for LFO pm calculations ...........*/
    unsigned int DT1val;    /*used for LFO pm calculations ...........*/

} OscilRec;

/* here's the virtual YM2151 */
struct YM2151 
{
    int status;     /*chip status (BUSY, IRQ Flags)*/

    OscilRec Oscils[32];    /*there are 32 operators in YM2151*/

    unsigned int LFOphase;  /*accumulated LFO phase (in indexes)    (fixed point)*/
    unsigned int LFOfrq;    /*LFO frequency (in indexes per sample) (fixed point)*/
    unsigned int PMD;   /*LFO Phase Modulation Depth*/
    unsigned int AMD;   /*LFO Amplitude Modulation Depth*/
    unsigned int LFOwave;   /*LFO waveform (0-saw, 1-square, 2-triangle, 3-random noise)*/
    unsigned int LFA;   /*current AM from LFO*/
    signed int LFP;     /*current PM from LFO*/

    unsigned int CT;    /*output control pins (bit7 CT2, bit6 CT1)*/
    unsigned int noise; /*noise register (bit 7 - noise enable, bits 4-0 - noise freq*/

    unsigned int IRQenable;     /*IRQ enable for timer B (bit 3) and timer A (bit 2)*/
#ifdef USE_MAME_TIMERS
    void *TimATimer,*TimBTimer; /* ASG 980324 -- added for tracking timers */
    double TimerATime[1024];    /*Timer A times for MAME*/
    double TimerBTime[256];     /*Timer B times for MAME*/
#else
    int TimA,TimB;          /*timer A,B enable (0-disabled)*/
    signed int TimAVal,TimBVal; /*current value of timer*/
    unsigned int TimerA[1024];  /*Timer A deltas*/
    unsigned int TimerB[256];   /*Timer B deltas*/
#endif
    unsigned int TimAIndex,TimBIndex;/*timers' indexes*/

    void (*handler)(void);      /*IRQ function handler*/
    void (*porthandler)(int offset,int data);   /*port write handler*/

    unsigned int freq[11*12*64];    /*11 octaves, 12 semitones, 64 'cents' */
    /*
    *   Frequency-deltas to get the closest frequency possible.
    *   There're 11 octaves because of DT2 (max 950 cents over base frequency)
    *   and LFO phase modulation (max 700 cents _over_ AND _below_ base frequency)
    *   Summary:   octave  explanation
    *              0       note code - LFO PM
    *              1       note code
    *              2       note code
    *              3       note code
    *              4       note code
    *              5       note code
    *              6       note code
    *              7       note code
    *              8       note code
    *              9       note code + DT2 + LFO PM
    *              10      note code + DT2 + LFO PM
    */

    signed int DT1freq[8*32];
    /*
    *   Frequency deltas for DT1. These deltas alter operator frequency
    *   after it has been taken from frequency-deltas table.
    */

    unsigned int LFOfreq[256];  /*frequency deltas for LFO*/

    unsigned int A_Time[64+32]; /*attack deltas (64 keycodes + 32 RKS's = 96)*/
    unsigned int D_Time[64+32]; /*decay  deltas (64 keycodes + 32 RKS's = 96)*/

    unsigned int clock; /* this will be passed from 2151intf.c */
    unsigned int sampfreq;  /* this will be passed from 2151intf.c */

    unsigned int need_refresh; /* 1 if emulator needs refreshing before calculating new samples*/
} ;


/*
** Initialize YM2151 emulator(s).
**
** 'num' is the number of virtual YM2151's to allocate
** 'clock' is the chip clock in Hz
** 'rate' is sampling rate
** 'sample_bits' is 8-bit/16-bit sample selector
*/
int YM2151Init(int num, int clock, int rate, int sample_bits);


/* shutdown the YM2151 emulators */
void YM2151Shutdown(void);


/* reset all chip registers for YM2151 number 'num'*/
void YM2151ResetChip(int num);

// save/load one ympsg chip status
void YM2151savestatus(FILE* f);
void YM2151loadstatus(FILE* f);

/*
** Generate samples for one of the YM2151's
**
** 'num' is the number of virtual YM2151
** '**buffers' is table of pointers to the buffers: left and right
** 'length' is the number of samples should be generated
*/
void YM2151UpdateOne(int num, SAMP **buffers, int length);


/* write 'v' to register 'r' on YM2151 chip number 'n'*/
void YM2151WriteReg(int n, int r, int v);


/* read status register on YM2151 chip number 'n'*/
int YM2151ReadStatus(int n);


/* set interrupt handler on YM2151 chip number 'n'*/
void YM2151SetIrqHandler(int n, void (*handler)(void));


/* set port write handler on YM2151 chip number 'n' */
void YM2151SetPortWriteHandler(int n, void (*handler)(int offset, int data));


#endif /* _H_YM2151_ */
